#ifndef __CMIOS9_H
#define __CMIOS9_H
/*
* Copyright (C) 2001, 2002, 2003, 2004  K. M. Indlekofer
*
* M.Indlekofer@fz-juelich.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#ifdef WIN32
/* e.g. VC++ */

#include "winlib_cmios9.h"

#else /* !WIN32 */
/* e.g. UNIX-like systems or cygwin */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <limits.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#endif /* !WIN32 */



#include "cmios9_io.h"



/* basic I/O: OS9, and MDR-DOS */
#ifndef BLOCKSIZE
#define BLOCKSIZE 256 /* blocksize in Bytes, fixed */
/* Note: for OS9/MDR-DOS, block-numbers are limited to 24 Bits (4GB) */
/*       => always fit into signed 32-Bit int! */
/*       Byte addresses however need unsigned 32-Bit int (or better: off_t)! */
#endif

/* basic I/O: QDOS */
/* Note: QDOS blocksize is 1/2 * BLOCKSIZE!!! */
#ifndef QDOS_BLOCKSIZE
#define QDOS_BLOCKSIZE 128 /* QDOS blocksize in Bytes, fixed */
/* Note: even now, 32-Bit int should be enough even for 24+1 Bits block-numbers (4GB)!!! */
#endif



/* OS9 data structures */

#define BMAPSTART 1 /* start of bitmap */

/* LSN0 */
struct lsn0_s{
	u_char dd_tot[3]; /* partition size, same as mdr_tot for MDR-DOS!!! */
	u_char dd_tks;
	u_char dd_map[2]; /* size of bitmap in Bytes, ==0 for MDR-DOS partition */
	u_char dd_bit[2]; /* size of cluster in blocks */
	u_char dd_dir[3];
	u_char dd_own[2];
	u_char dd_att;
	u_char dd_dsk[2];
	u_char dd_fmt;
	u_char dd_spt[2];
	u_char dd_res[2];
	u_char dd_bt[3];
	u_char dd_bsz[2];
	u_char dd_date[5];
#define LSN0_NAMELEN 32
	char dd_name[LSN0_NAMELEN]; /* first char starts later in MDR-DOS partition */
	char dummy1[49];
	u_char dd_ptabstart[3];
	u_char dd_pnr;
	char dummy2[2];
	u_char dd_qdosstart[3]; /* QDOS image partitions */
	u_char dd_qdosnr[2];
	u_char dd_cmistart[3];  /* CMI image partitions */
	u_char dd_cminr[2];
	char dummy3[128];
};

/* partition file entry */
struct parttabent_s{
	u_char index;
	u_char startsec[3];
	u_char size[3];
	char dummy[9];
};
#define PARTTAB_NR 16 /* max. number of partitions */

/* standard k-partition */
#define KPART_STD_PARTSIZE 0x10cc
#define KPART_STD_FSYSSIZE 0x10cc
#define KPART_STD_BITMSIZE 0x10d0 /* FSYSSIZE according to clustersize*bitmapsize */

/* segment descriptor */
struct seg_s{
	u_char addr[3]; /* in BLOCKSIZE */
	u_char size[2]; /* in BLOCKSIZE */
};
#define SEG_S_SIZEMAX 0xffff /* in BLOCKSIZE */

/* file descriptor */
struct filedes_s{
	u_char fd_att;
	u_char fd_own[2];
	u_char fd_dat[5];
	u_char fd_link;
	u_char fd_fsize[4];
	u_char fd_creat[3];
#define FILEDES_SEGNR 48
	struct seg_s fd_seg[FILEDES_SEGNR];
};

/* file and device names are case-insensitive when compared */
#define FILENAMELEN 29 /* max. file name length */

/* directory entry */
struct dirent_s{
	char dir_name[FILENAMELEN];
	u_char dir_addr[3];
};
#define DIRBLKS 8 /* XXX directory file allocation granularity */

/* attributes */
#define ATT_D  0x80 /* directory */
#define ATT_S  0x40 /* non-sharable */
#define ATT_PE 0x20 /* public execute */
#define ATT_PW 0x10 /* public write */
#define ATT_PR 0x08 /* public read */
#define ATT_E  0x04 /* owner execute */
#define ATT_W  0x02 /* owner write */
#define ATT_R  0x01 /* owner read */



/* MDR-DOS data structures */

/* sizes and starts in BLOCKSIZE! */
#define MDR_CLUSTERSIZE  0x001000 /* granularity */
#define MDR_BMAPSTART    0x000004 /* start of Bitmap */
#define MDR_BMAPSIZE     0x000002 /* Bitmap size */
#define MDR_DIRSTART     0x000006 /* start of directory */
#define MDR_DIRENTRYSIZE 0x000028 /* file entry in directory */

/* LSN0 */
struct mdr_lsn0_s{
	u_char mdr_tot[3];  /* same as dd_tot for OS9!!! */
	u_char mdr_clsz;    /* cluster size in MDR_CLUSTERSIZE blocks! */
	u_char zero1[2];    /* must be zero: !=0 for OS9 partition */
	u_char mdr_clnr[2]; /* total number of clusters */
	u_char zero2[2];    /* must be zero */
	u_char mdr_dfnr[2]; /* total number of file entries in directory */
	u_char zero3[2];    /* must be zero */
	u_char mdr_fnr[2];  /* last valid file entry number+1 */
	u_char zero4[16];   /* must be zero */
#define MDR_LSN0_NAMELEN 32
	char mdr_name[MDR_LSN0_NAMELEN];  /* name (null-terminated) */
	u_char zero5[192];  /* must be zero */
};

#define MDR_LSN0_P(p) ((struct mdr_lsn0_s *)(p))
/* Note: mdr_lsn0_s fits into lsn0_s!!! */

/* directory entry */
#define MDR_FILECLNR 0xf00  /* max. number of clusters in file */

/* should be MDR_DIRENTRYSIZE*BLOCKSIZE */
#define MDR_DIRENT_NAMELEN 32
struct mdr_dirent_s{
	u_char dummy1[512];
	char name[MDR_DIRENT_NAMELEN]; /* file name (null-terminated), name[0]==0 if not used  */
#define MDR_MDOSTYPE "MDOS"
	char type[4];    /* type */
	u_char size[4];  /* file size in Bytes */
	u_char dummy2[8];
	u_char back[6];  /* creation/backup date: yy-mm-dd hh:mm:ss */
	u_char dummy3[10];
	u_char modif[6]; /* modification date: yy-mm-dd hh:mm:ss */
	u_char dummy4[10];
	u_char dummy5[0x7b0];
	u_char clist[MDR_FILECLNR][2]; /* cluster list */
};



/* QDOS data structures (derived from MDOS) */
/* Note: partitions are for floppy images, supported by "QDOS Tackiton driver" */
/* Note: system floppy is in CMI Series IIx QDOS-6809 low-level format */
/*       sound floppy is in CMI Series I/II QDOS-6800 low-level format */

#define QDOS_PBLKS (2*26*77) /* number QDOS blocks within QDOS partition */
							 /* Note: comes from 2 heads, 26 sectors/track, 77 tracks */
#define QDOS_PSIZE (QDOS_PBLKS*QDOS_BLOCKSIZE) /* QDOS partition size in Bytes */
#define QDOS_CSIZE 4 /* fixed QDOS cluster size in QDOS blocks */

/* QDOS disk ID (block 0) */
/* should be QDOS_BLOCKSIZE */
struct qdos_did_s{
	char did_id[8];  /* disk ID (unused chars filled with space) */
	char did_vn[2];  /* version number */
	char did_rn[2];  /* revision number */
	char did_dt[6];  /* date: DD-MM-YY */
	char did_nm[20]; /* user name (unused chars filled with space) */
#define QDOS_DID_RBNR 7
	u_char did_rb[QDOS_DID_RBNR][2]; /* overlay RIB addresses in QDOS blocks */
	u_char did_btsec[2]; /* bootblock: start sector in QDOS blocks */
						 /* Note: typ. QDOS_BTSEC */
	u_char did_btsiz[2]; /* bootblock: size in QDOS blocks */
						 /* Note: typ. QDOS_BTSIZ */
	u_char did_btloa[2]; /* bootblock: load address in Bytes */
						 /* Note: typ. QDOS_BTLOA */
	char did_slinflag; /* start line flag */
	char did_slin[69]; /* start line (shell command, executed at boot time) */
};

#define QDOS_BTSEC       0x0017 /* default bootblock in QDOS blocks */
								/* Note: actual value in did_btsec */
#define QDOS_BTSIZ       0x0002 /* default bootblock load size in QDOS blocks */
								/* Note: default loads QDOS_BTSEC and QDOS_DOSRIB */
								/* Note: actual value in did_btsiz */
#define QDOS_BTLOA       0x0080 /* default bootblock load address in Bytes */
								/* Note: actual value in did_btloa */
#define QDOS_DOSRIB      0x0018 /* default RIB of xxxDOS.SY in QDOS blocks */
								/* Note: typ. loaded together with QDOS_BTSEC */
#define QDOS_FSYSSTART   0x0018 /* start of file-system area in QDOS blocks */
								/* Note: typ. xxxDOS.SY starts here */

#define QDOS_LSN0_P(p) ((struct qdos_did_s *)(p))
/* Note: qdos_did_s fits into lsn0_s!!! */

/* cluster allocation table (CAT): 1 block */
#define QDOS_CATSTART     0x0001 /* in QDOS blocks */
/* lockout cluster allocation table (LCAT): 1 block */
#define QDOS_LCATSTART    0x0002 /* in QDOS blocks */

/* directory */
#define QDOS_DIRSTART     0x0003 /* in QDOS blocks */
#define QDOS_DIRSIZE      0x0014 /* in QDOS blocks */
#define QDOS_DIRENTRYSIZE 0x0010 /* in Bytes */
#define QDOS_DIRENTRYNR   ((QDOS_DIRSIZE*QDOS_BLOCKSIZE)/QDOS_DIRENTRYSIZE)
                          /* number of files in directory */
/* QDOS directory entry */
/* should be QDOS_DIRENTRYSIZE */
struct qdos_dir_s{
#define QDOS_DIR_VACANT  0x00 /* first char in dir_nm if first vacant entry */
#define QDOS_DIR_DELETED ((char)0xff) /* first char in dir_nm if deleted entry */
									  /* Note: typ. also second char */
	char dir_nm[8]; /* name (unused chars filled with space) */
	char dir_sx[2]; /* suffix (unused chars filled with space) */
	u_char dir_rb[2]; /* RIB address in QDOS blocks */
	u_char dir_at[2]; /* attributes */
	u_char dir_nu[2]; /* not used area */
};

/* attributes */
/* Byte 0: */
#define QDOS_ATT_FLM 0xf8 /* flag mask */
#define QDOS_ATT_WRT 0x80 /* write protect */
#define QDOS_ATT_DEL 0x40 /* delete protect */
#define QDOS_ATT_SYS 0x20 /* system file */
#define QDOS_ATT_CON 0x10 /* contiguous allocation */
#define QDOS_ATT_NCS 0x08 /* non-compressed spaces */
#define QDOS_ATT_FMM 0x07 /* format mask */
#define QDOS_ATT_FMU 0x00 /* user-defined format */
#define QDOS_ATT_FMD 0x01 /* default object record format */
#define QDOS_ATT_FML 0x02 /* binary load format */
#define QDOS_ATT_FMB 0x03 /* binary record format */
#define QDOS_ATT_FMA 0x05 /* ASCII record format */
#define QDOS_ATT_FMC 0x07 /* ASCII-converted-binary record format */
/* Byte 1: */
/* XXX */

/* QDOS "retrieval information block" (RIB) */
/* should be QDOS_BLOCKSIZE */
/* Note: first block of file allocation (in rib_seg) is used for RIB!!! */
struct qdos_rib_s{
#define QDOS_RIB_SDWEND    0x8000 /* terminator (end flag) */
#define QDOS_RIB_SDWTOTSIZ 0x7fff /* if end flag: mask for file data size-1 in QDOS blocks */
#define QDOS_RIB_SDWSIZ    0x7c00 /* mask for segment size-1 in QDOS clusters */
#define QDOS_RIB_SDWSIZMAX 0x0020 /* max. number of clusters per segment */
#define QDOS_RIB_SDWOFF    0x03ff /* mask for segment start in QDOS clusters */
#define QDOS_RIB_SDWNR     58
	u_char rib_sdw[QDOS_RIB_SDWNR][2]; /* segment descriptor words (SDW) */
	u_char zero0;
	u_char rib_lb; /* number of Bytes in last sector(==block) */
				   /* rib_lb==0 means full block (QDOS_BLOCKSIZE) */
				   /* MDOS requires rib_lb%8==0 */
	u_char rib_sl[2]; /* number of sectors(==blocks) to load */
	u_char rib_la[2]; /* memory load address */
	u_char rib_sa[2]; /* start execution address */
	u_char zero1[4];
};



/* personal data structures */

/* partition */
struct part_s{
	/* general */
#define PARTTYPE_UNDEF  0 /* undefined */
#define PARTTYPE_OS9MDR 1 /* OS9/MDR-DOS */
#define PARTTYPE_QDOS   2 /* QDOS */
	u_int type; /* partition/filesystem type */
	u_int master; /* !=0 if OS9/MDR-DOS master partition */
	u_int mdr; /* !=0 if MDR-DOS partition */
	u_int blksiz; /* blocksize of partition in Bytes */
	u_int startsec; /* in blksiz!!! */
	u_int psize;   /* size according to partition table entry in blksiz!!! */
	               /* or totsize */
	u_int totsize; /* size (dd_tot) according to LSN0 of partition in blksiz!!! */
	               /* if ==0: partition marked invalid! */
	struct lsn0_s *lsn0; /* pointer to LSN0 of this partition */
	int fdes; /* file descriptor of this partition */
	/* OS9 */
	u_int bitmapsize; /* in Bytes */
	u_int clustersize;
	u_int filesyssize;
	u_int rootdirsec;
	u_int bootsec;
	u_int bootsize;
	u_int ptabsec;
	u_int pnr;
	/* MDR-DOS */
	u_int mdr_clustersize; /* now in BLOCKSIZE!!! */
	u_int mdr_clusternr;
	u_int mdr_filesyssize;
	u_int mdr_dfilenr;
	u_int mdr_filenr;
	/* QDOS */
	/* XXX */
	/* general */
	char devname[FILENAMELEN+1];
};

/* "QDOS file descriptor" */
/* Note: RIB not enough: need dir_at as well!!! */
struct qdos_fdes_s{
	struct qdos_dir_s dir;
	struct qdos_rib_s rib;
};

/* general */
/* generic file descriptor */
union genfdes_u{
	struct filedes_s os9;       /* OS9 */
	struct mdr_dirent_s mdrdos; /* MDR-DOS */
	struct qdos_fdes_s qdos;    /* QDOS */
};

/* general */
/* generic file */
struct genfile_s{
	union genfdes_u fdes;
	struct part_s *part;
	u_int startsec;
	u_int fdessec;
	u_int size;
};



/* general */
/* current partition I/O */
#ifndef READBLOCK_CURPART
#define READBLOCK_CURPART(a,b,c) readblock(curpart->fdes,a,b,c,curpart->startsec,curpart->totsize,curpart->blksiz)
#endif
#ifndef WRITEBLOCK_CURPART
#define WRITEBLOCK_CURPART(a,b,c) writeblock(curpart->fdes,a,b,c,curpart->startsec,curpart->totsize,curpart->blksiz)
#endif



/* general */
#define PARTITION_NR 128
extern struct lsn0_s lsn0[PARTITION_NR];
extern struct part_s part[PARTITION_NR];
extern u_int partnr; /* number of used partitions */

/* current partition */
extern struct part_s *curpart;
extern u_int curowner; /* who we are */



/* OS9 */
/* current directory */
extern u_int curdirsec;

/* general */
#define DIRLEVELNR 8 /* just for display-purposes, no OS9-specific limitation!!! */
extern int dirlevel; /* 0: root, level index to dirlist=dirlevel-1 */
extern char dirlist[DIRLEVELNR][FILENAMELEN];

/* general */
/* device aliases */
struct devalias_s{
	char *alias;
	char *equ;
};

extern struct devalias_s devalias[];



extern void translatename(char *name,char *buf);
extern void translatebackname(char *name,char *buf,u_int size);
extern void printname(char *name);
extern void printnstr(char *str,int maxn);
extern void qdos_makefname(char *in,char *out);
extern int lsn0label(char *name);
extern int qdos_didlabel(char *id,char *vnrn,char *date,char *user);
extern void setdate(struct filedes_s *fd,int cre);
extern void lsn0setdate(struct lsn0_s *pp);
extern void mdr_setdate(struct mdr_dirent_s *fd,int cre);
extern int setseg(struct seg_s *seglist,u_int start,u_int size);
extern int getseg(struct seg_s *seglist,u_int *startp,u_int *sizep);
extern u_int countseg(struct seg_s *seglist,u_int segnr,u_int *contblp);
extern u_int mdr_countclu(struct mdr_dirent_s *direntp,u_int clunr,u_int *contclp);
extern int qdos_setseg(struct qdos_rib_s *ribp,u_int seg,u_int start,u_int size);
extern int qdos_getseg(struct qdos_rib_s *ribp,u_int seg,u_int *startp,u_int *sizep);
extern u_int qdos_countseg(struct qdos_rib_s *ribp,u_int segnr,u_int *contclp);
extern u_int qdos_fsize(struct qdos_rib_s *ribp);
extern int qdos_setfsize(struct qdos_rib_s *ribp,u_int fsize);
extern int bitmaptranslate(u_int sec,u_int *bmapbitp,u_int *bmapbytep,u_int *bmapblkp);
extern int mdr_bitmaptranslate(u_int sec,u_int *bmapbitp,u_int *bmapbytep,u_int *bmapblkp);
extern int qdos_bitmaptranslate(u_int sec,u_int *bmapbitp,u_int *bmapbytep);
extern int markbitmap2(u_int startblk,u_int size,int markused,int check);
extern int markbitmap(struct seg_s *seglist,u_int segnr,int markused);
extern int mdr_markbitmap2(u_int startblk,u_int size,int markused,int check);
extern int mdr_markbitmap(struct mdr_dirent_s *direntp,u_int fclunr,int markused);
extern int qdos_markbitmap2(u_int startblk,u_int size,int markused,int check,int lcat);
extern int qdos_markbitmap(struct qdos_rib_s *ribp,u_int segnr,int markused);
extern int initbitmap(void);
extern int mdr_initbitmap(void);
extern int qdos_initbitmap(int single,int newlcat);
extern int newfs(u_int totsize,u_int clustersize,u_int filesyssize,int floppyflag);
extern int mdr_newfs(u_int totsize,u_int clustersize,u_int dfilenr);
extern int qdos_newfs(int single,int newlcat);
extern int findfree(u_int seekstart,u_int minsize,
					u_int *segstartp,u_int *segsizep,
					u_int *lsegstartp,u_int *lsegsizep,u_int *segnrp,
					struct seg_s *skiplist,u_int skipnr,int fullscan);
extern int mdr_findfree(u_int seekstart,u_int minsize,
						u_int *segstartp,u_int *segsizep,
						u_int *lsegstartp,u_int *lsegsizep,u_int *segnrp,
						struct mdr_dirent_s *skipdirentp,u_int skipnr,int fullscan);
extern int qdos_findfree(u_int seekstart,u_int minsize,
						 u_int *segstartp,u_int *segsizep,
						 u_int *lsegstartp,u_int *lsegsizep,u_int *segnrp,
						 struct qdos_rib_s *skipribp,u_int skipnr,int fullscan);
extern int findfreelist(struct seg_s *seglist,u_int segnr,
						u_int startblk,u_int blknr);
extern int mdr_findfreelist(struct mdr_dirent_s *direntp,u_int startfclu,u_int fclunr,
							u_int startblk,u_int blknr);
extern int qdos_findfreelist(struct qdos_rib_s *ribp,u_int startseg,u_int segnr,
							 u_int startblk,u_int blknr);
extern int appendfreelist(struct seg_s *seglist,u_int segnr,u_int blknr);
extern int mdr_appendfreelist(struct mdr_dirent_s *direntp,u_int startfclu,u_int fclunr,u_int blknr);
extern int qdos_appendfreelist(struct qdos_rib_s *ribp,u_int startseg,u_int segnr,u_int blknr);
extern void qdos_printatt0(u_char att);
extern void printatt(u_char att);
extern int setattr(struct filedes_s *fdes,char *str);
extern int qdos_setattr(struct qdos_rib_s *ribp,u_int fnr,void *indirp,char *attstr);
extern u_int printfiledes(struct part_s *pp,struct filedes_s *fdes,int verbose);
extern void mdr_printdirent(struct mdr_dirent_s *direntp);
extern void qdos_printfdes(	struct qdos_dir_s *dirp,struct qdos_rib_s *ribp);
extern int filelseek(union genfdes_u *fp,u_int lblk,u_int *contblksp);
extern int fileread(union genfdes_u *fp,void *buf,u_int lblkstart,u_int lblknr);
extern int filewrite(union genfdes_u *fp,void *buf,u_int lblkstart,u_int lblknr);
extern int filedel(u_int fdessec);
extern int mdr_filedel(struct mdr_dirent_s *direntp);
extern int qdos_filedel(struct qdos_rib_s *qrib);
extern int filecreat(struct filedes_s *fdes,u_int blknr,int extend);
extern int mdr_filecreat(struct mdr_dirent_s *direntp,u_int blknr,int extend);
extern int qdos_filecreat(struct qdos_rib_s *ribp,u_int blknr,int extend);
extern int getdir(u_int newdirsec,struct filedes_s *curdirdesp,u_int *curdirsizep);
extern int findfile(char *name,int countfree,u_int *freeentp);
extern int mdr_findfile(char *name);
extern int qdos_findfile(char *name);
extern int deldirent(char *name,int dir);
extern int creatdirent(char *name,u_int startsec);
extern int changedir2(char *name);
extern int delfile(char *name,int dir);
extern int mdr_updatefilenr(void);
extern int mdr_delfile(char *name);
extern int qdos_delfile(char *name);
extern int creatfile(char *name,u_int bytesize,int dir,struct filedes_s *tmplt);
extern int mdr_creatfile(char *name,u_int bytesize,struct mdr_dirent_s *tmplt);
extern int qdos_creatfile(char *name,u_int bytesize,struct qdos_fdes_s *tmplt);
extern int readlsn0(void);
extern void printlsn0(void);
extern int partsetup(int fdes,u_int devid,int forceqdosnr);
extern int partmodif(struct parttabent_s *parttab,u_int index,u_int size,u_int startsec,int clearall);
extern struct part_s *partchange2(char *name,int check);
extern int printpartinfo(char *part,u_int seekstart,u_int minsize,int mode);
extern void printpartlist(int mode);
extern int findfile3(char *path,u_int *startsecp,struct part_s **partp);
extern int dirlistupdate3(char *path);
extern void printdirlist(void);
extern int changedir3(char *path,int update);
extern int creatfile3(char *path,u_int bytesize,int dir,struct genfile_s *tmplt);
extern int readfiledes3(char *path,union genfdes_u *fp,u_int *typep,u_int *mdrp,
						u_int *startsecp,struct part_s **partp,u_int *fdessecp,
						int pchg,u_int *sizep,int touch);
extern int extendfile3(char *path,u_int bytesize);
extern int lsn0reset(void);
extern int lsn0bootfile(char *path,int bsz);
extern int lsn0partfile(char *path);
extern int qdos_ovfile(char *path,u_int ovnr);
extern int qdos_setboot(u_int btsec,u_int btsiz,u_int btloa);
extern int listdirtree(int (*recfct)(char *name,int rec,int arg),int arg);
extern int mdr_listdir(void);
extern int qdos_listdir(void);
extern int listdir(int (*recfct)(char *name,int rec,int arg),int arg);
extern int printdir3(char *path,int recursive,int tree);
extern int openfile3(char *path,struct genfile_s *fp,int touch);
extern int readfile4(struct genfile_s *fp,char *buf,u_int bytestart,u_int bytenr);
extern int writefile4(struct genfile_s *fp,char *buf,u_int bytestart,u_int bytenr);
extern int movefile3(char *path1,char *path2,int link);
extern int delfile3 (char *path,int dir);
extern int filesetattr3(char *path,char *str,int own);



#endif /* !__CMIOS9_H */
